@@ -14,13 +14,13 @@ my $case_fold = $] >= 5.016		    ? eval 'sub { return CORE::fc $_[0] }'
 	      : $INC{'Unicode/CaseFold.pm'} ? eval 'sub { return Unicode:CaseFold::fc $_[0] }'
 	      : 			      eval 'sub { return lc $_[0] }';
 
-our $VERSION = '0.45';
+our $VERSION = '0.47';
 
 =head1 NAME
 
 HTTP::Validate - validate and clean HTTP parameter values according to a set of rules
 
-Version 0.45
+Version 0.47
 
 =head1 DESCRIPTION
 
@@ -387,13 +387,14 @@ To match an empty parameter value, include a string that looks like
 well-known types 'html', 'json', 'xml', 'txt' or 'csv', unless you wish to
 override the default given by this module.
 
-=head2 Other keys
+=head2 Rule attributes
 
-Any rule definition may also include one or more of the following keys:
+Any rule definition may also include one or more of the following attributes,
+specified as key/value pairs in the rule hash:
 
 =head3 errmsg
 
-This key specifies the error message to be returned if the rule fails,
+This attribute specifies the error message to be returned if the rule fails,
 overriding the default message.  For example:
 
     define_ruleset( 'specifier' => 
@@ -405,17 +406,15 @@ overriding the default message.  For example:
           errmsg => "you must specify either of the parameters 'name' or 'id'" });
 
 Error messages may include any of the following placeholders: C<{param}>,
-C<{value}>.  When used in a parameter rule these are replaced respectively by
-the parameter name and original parameter value(s), single-quoted.  When used
-with other rules, {param} is replaced by the full list of relevant parameters
-or ruleset names, quoted and separated by commas.  This feature allows you to
+C<{value}>.  These are replaced respectively by the relevant parameter name(s)
+and original parameter value(s), single-quoted.  This feature allows you to
 define messages that quote the actual parameter values presented in the
 request, as well as to define common messages and use them with multiple
 rules.
 
 =head3 warn
 
-This key causes a warning to be generated rather than an error if the
+This attribute causes a warning to be generated rather than an error if the
 rule fails.  Unlike errors, warnings do not cause a request to be rejected.
 At the end of the validation process, the list of generated warnings can be
 retrieved by using the L</warnings> method of the result object.
@@ -424,9 +423,14 @@ If the value of this key is 1, then what would otherwise be the error
 message will be used as the warning message.  Otherwise, the specified string
 will be used as the warning message.
 
+For parameter rules, this attribute affects only errors resulting from
+validation of the parameter values.  Other error conditions (i.e. multiple
+parameter values without the L</multiple> attribute) continue to be reported
+as errors.
+
 =head3 key
 
-The key 'key' specifies the name under which any information generated by
+The attribute 'key' specifies the name under which any information generated by
 the rule will be saved. For a parameter rule, the cleaned value will be saved
 under this name.  For all rules, any generated warnings or errors will be
 stored under the specified name instead of the parameter name or rule number.
@@ -438,35 +442,35 @@ L</param>, L</optional> or L</mandatory>:
 
 =head3 valid
 
-This key specifies the domain of acceptable values for the parameter.  The
+This attribute specifies the domain of acceptable values for the parameter.  The
 value must be either a single code reference or a list of them.  You can
 either select from the list of L<built-in validator functions|/VALIDATORS>
 included with this module, or provide your own.
 
-If the parameter named by this rule is present, it must pass at least one of
-the specified validators or else an error message will be generated.  If
-multiple validators are given, then the error message returned will be the one
-generated by the last validator in the list.  This can be overridden by using
-the L</errmsg> key.
+If the parameter named by this rule is present, its value must pass at least
+one of the specified validators or else an error message will be generated.
+If multiple validators are given, then the error message returned will be the
+one generated by the last validator in the list.  This can be overridden by
+using the L</errmsg> key.
 
 =head3 multiple
 
-This key specifies that the parameter may appear multiple times in the
+This attribute specifies that the parameter may appear multiple times in the
 request.  Without this directive, multiple values for the same parameter will
 generate an error.  For example:
 
     define_ruleset( 'identifiers' => 
 	{ param => 'id', valid => POS_VALUE, multiple => 1 });
 
-If this directive is used, then the cleaned value of the parameter will be an
-array ref if at least one valid value was found and I<undef> otherwise.  If
-you wish a request to be considered valid even if some of the values fail the
-validator, then either use the L</list> key instead or include a
-L</warn> key as well.
+If this attribute is present with a true value, then the cleaned value of the
+parameter will be an array ref if at least one valid value was found and
+I<undef> otherwise.  If you wish a request to be considered valid even if some
+of the values fail the validator, then either use the L</list> attribute instead or
+include a L</warn> key as well.
 
 =head3 split
 
-This directive has the same effect as L</multiple>, and in addition causes
+This attribute has the same effect as L</multiple>, and in addition causes
 each parameter value string to be split (L<perlfunc/split>) as indicated by the
 value of the directive.  If this value is a string, then it will be compiled
 into a regexp preceded and followed by C<\s*>.  So in the
@@ -490,22 +494,40 @@ regexp quoted with L<qr> instead.
 
 =head3 list
 
-This directive has the same effect as L</split>, but generates warnings
+This attribute has the same effect as L</split>, but generates warnings
 instead of error messages when invalid values are encountered (as if 
 C<< warn => 1 >> was also specified).  The resulting cleaned value will be a
 listref containing any values which pass the validator, or I<undef> if no
-valid values were found.  See also L</warn>.
+valid values were found.  See also L</warn> and L</bad_value>.
+
+=head3 bad_value
+
+This attribute can be useful in conjunction with L</list>.  If one or more
+values are given for the parameter but none of them are valid, this attribute
+comes into effect.  If the value of this attribute is C<ERROR>, then the
+validation will fail with an appropriate error message.  Otherwise, this will
+be used as the value of the parameter.  It is recommended that you set the
+value to something outside of the valid range, i.e. C<-1> for a C<POS_VALUE>
+parameter.
+
+Using this attribute allows you to easily distinguish between the case when
+the parameter appears with an empty value (or not at all, which is considered
+equivalent) vs. when the parameter appears with one or more invalid values and
+no good ones.
 
 =head3 alias
 
-This directive specifies one or more aliases for the parameter name (use a
+This attribute specifies one or more aliases for the parameter name (use a
 listref for multiple aliases).  These names may be used interchangeably in
 requests, but any request that contains more than one of them will be rejected
-with an appropriate error message unless L</multiple> is also specified.
+with an appropriate error message unless L</multiple> is also specified.  The
+parameter value and any error or warning messages will be reported under the
+main parameter name for this rule, no matter which alias is used in the
+request.
 
 =head3 clean
 
-This directive specifies a subroutine which will be used to modify the
+This attribute specifies a subroutine which will be used to modify the
 parameter values.  This routine will be called with the raw value of the
 parameter as its only argument, once for each value if multiple values are
 allowed.  The resulting values will be stored as the "cleaned" values.  The
@@ -515,11 +537,11 @@ uppercase, lowercase, or L<fold case|Unicode::CaseFold> respectively.
 
 =head3 default
 
-This directive specifies a default value for the parameter, which will be
-reported if no value is specified in the request.  If the rule also includes a
-validator, the specified default value will be passed to it and the resulting
-cleaned value, if any, will be used.  An exception will be thrown at the time
-of rule definition if the default value does not pass the validator.
+This attribute specifies a default value for the parameter, which will be
+reported if the parameter is not present in the request or if it is present
+with an empty value.  If the rule also includes a validator and/or a cleaner,
+the specified default value will be passed to it when the ruleset is defined.
+An exception will be thrown if the default value does not pass the validator.
 
 =head2 Documentation
 
@@ -631,14 +653,30 @@ my ($DEFAULT_INSTANCE) = bless { RULESETS => {}, SETTINGS => {} };
 # Known media types are defined here
 
 my (%MEDIA_TYPE) = 
-    ('html' => 'text/html',
-     'xml' => 'text/xml',
-     'txt' => 'text/plain',
-     'tsv' => 'text/tab-separated-values',
-     'csv' => 'text/csv',
-     'json' => 'application/json',
-    );
+   ('html' => 'text/html',
+    'xml' => 'text/xml',
+    'txt' => 'text/plain',
+    'tsv' => 'text/tab-separated-values',
+    'csv' => 'text/csv',
+    'json' => 'application/json',
+   );
 
+# Default error messages
+
+my (%ERROR_MSG) = 
+   ('ERR_INVALID' => "the value of parameter {param} is invalid (was {value})",
+    'ERR_BAD_VALUES' => "no valid values were specified for {param} (found {value})",
+    'ERR_MULT_NAMES' => "you may only include one of {param}",
+    'ERR_MULT_VALUES' => "you may only specify one value for {param}: found {value}",
+    'ERR_MANDATORY' => "you must specify a value for {param}",
+    'ERR_TOGETHER' => "you must specify {param} together or not at all",
+    'ERR_AT_MOST' => "you may not specify more than one of {param}",
+    'ERR_REQ_SINGLE' => "you must specify the parameter {param}",
+    'ERR_REQ_MULT' => "you must specify at least one of the parameters {param}",
+    'ERR_REQ_ONE' => "you may not include parameters from more than one of these groups: {param}",
+    'ERR_MEDIA_TYPE' => "you must specify a media type, from the following list: {value}",
+    'ERR_DEFAULT' => "parameter value error: {param}",
+   );
 
 =head1 INTERFACE
 
@@ -850,6 +888,11 @@ sub validation_settings {
 	    $self->{SETTINGS}{ignore_unrecognized} = $value ? 1 : 0;
 	}
 	
+	elsif ( $ERROR_MSG{$key} )
+	{
+	    $self->{SETTINGS}{$key} = $value;
+	}
+	
 	else
 	{
 	    croak "unrecognized setting: '$key'";
@@ -1012,8 +1055,8 @@ my %DIRECTIVE = ( 'param' => 2, 'optional' => 2, 'mandatory' => 2,
 		  'together' => 2, 'at_most_one' => 2, 'ignore' => 2,
 		  'require' => 2, 'allow' => 2, 'require_one' => 2,
 		  'require_any' => 2, 'allow_one' => 2, 'content_type' => 2,
-		  'valid' => 1, 'bad_value' => 1, 'clean' => 1,
-		  'multiple' => 1, 'split' => 1, 'list' => 1,
+		  'valid' => 1, 'clean' => 1, 
+		  'multiple' => 1, 'split' => 1, 'list' => 1, 'bad_value' => 1, 
 		  'error' => 1, 'errmsg' => 1, 'warn' => 1, 
 		  'alias' => 1, 'key' => 1, 'default' => 1);
 
@@ -1152,11 +1195,11 @@ sub add_rules {
     KEY:
 	foreach my $key (keys %$rule)
 	{
-	    croak "unknown key '$key' found in rule" unless $DIRECTIVE{$key};
+	    croak "unknown attribute '$key' found in rule" unless $DIRECTIVE{$key} || $ERROR_MSG{$key};
 	    
-	    if ( $DIRECTIVE{$key} == 2 )
+	    if ( defined $DIRECTIVE{$key} && $DIRECTIVE{$key} == 2 )
 	    {
-		croak "a rule definition cannot contain the keys '$key' and '$type' together, because they indicate different rule types"
+		croak "a rule definition cannot contain the attributes '$key' and '$type' together, because they indicate different rule types"
 		    if $type;
 		$type = $key;
 		$rr->{$type} = $rule->{$type};
@@ -1172,13 +1215,13 @@ sub add_rules {
 	    
 	    if ( $key eq 'valid' )
 	    {
-		croak "the key 'valid' is only allowed with parameter rules"
+		croak "the attribute 'valid' is only allowed with parameter rules"
 		    unless $CATEGORY{$type} eq 'param' || $type eq 'content_type';
 	    }
 	    
 	    elsif ( $key eq 'alias' )
 	    {
-		croak "the key 'alias' is only allowed with parameter rules"
+		croak "the attribute 'alias' is only allowed with parameter rules"
 		    unless $CATEGORY{$type} eq 'param';
 		
 		croak "the value of 'alias' must be a string or a list ref"
@@ -1189,7 +1232,7 @@ sub add_rules {
 	    
 	    elsif ( $key eq 'clean' )
 	    {
-		croak "they key 'clean' is only allowed with parameter rules"
+		croak "they attribute 'clean' is only allowed with parameter rules"
 		    unless $CATEGORY{$type} eq 'param';
 		
 		$rr->{cleaner} = $CLEANER_DEF{$value} || $value;
@@ -1200,7 +1243,7 @@ sub add_rules {
 	    
 	    elsif ( $key eq 'default' )
 	    {
-		croak "the key 'default' is only allowed with parameter rules"
+		croak "the attribute 'default' is only allowed with parameter rules"
 		    unless $CATEGORY{$type} eq 'param';
 		
 		$rr->{default} = $value;
@@ -1208,7 +1251,7 @@ sub add_rules {
 	    
 	    elsif ( $key eq 'split' || $key eq 'list' )
 	    {
-		croak "the key '$key' is only allowed with parameter rules"
+		croak "the attribute '$key' is only allowed with parameter rules"
 		    unless $CATEGORY{$type} eq 'param';
 		
 		croak "the value of '$key' must be a string or a regexp"
@@ -1231,7 +1274,7 @@ sub add_rules {
 	    
 	    elsif ( $key eq 'error' || $key eq 'errmsg' )
 	    {
-		$rr->{errmsg} = $rr->{$key};
+		$rr->{errmsg} = $value;
 	    }
 	    
 	    elsif ( $key ne $type )
@@ -1343,8 +1386,8 @@ sub add_rules {
 	    
 	    foreach my $arg (@params)
 	    {
-		croak "parameter '$arg' was not defined" unless defined $rs->{params}{$arg} ||
-		    $type eq 'ignore';
+		# croak "parameter '$arg' was not defined" unless defined
+		# $rs->{params}{$arg} || $type eq 'ignore';
 		push @{$rr->{param}}, $arg;
 	    }
 	    
@@ -1768,11 +1811,14 @@ sub new_execution {
     
     # Now create a new validation record
     
+    my %settings = %{$self->{SETTINGS}};
+    
     my $vr = { raw => $unpacked_params,	# the raw parameters and values
 	       clean => {},		# the cleaned parameters and values
 	       context => $context,	# context for the validators to use
 	       ps => {},		# the status (failed=0, passed=1, ignored=undef) of each parameter
 	       rs => {},		# the status (checked=1, fulfilled=2) of each ruleset
+	       settings => \%settings,	# a copy of our current settings
 	     };
     
     return bless $vr, 'HTTP::Validate::Progress';
@@ -1834,9 +1880,8 @@ sub execute_validation {
     if ( $vr->{rs}{$ruleset_name} != 2 )
     {
 	my @names = @{$self->{RULESETS}{$ruleset_name}{fulfill_order}};
-	my $message = @names == 1 ? "you must include the parameter {param}" :
-	    "you must include at least one of the parameters {param}";
-	add_error($vr, { key => $ruleset_name }, $message, { param => \@names });
+	my $msg = @names == 1 ? 'ERR_REQ_SINGLE': 'ERR_REQ_MULT';
+	add_error($vr, { key => $ruleset_name }, $msg, { param => \@names });
     }
     
     # Create an object to hold the result of this function.
@@ -1845,9 +1890,14 @@ sub execute_validation {
     
     # Add the clean-value hash and the raw-value hash
     
-    $result->{values} = $vr->{clean};
+    $result->{clean} = $vr->{clean};
     $result->{raw} = $vr->{raw};
     
+    # Put the clean-value hash under the old name, for backward compatibility
+    # (it will be eventually removed).
+    
+    $result->{values} = $vr->{clean};
+    
     # Add the content type, if one was specified.
     
     $result->{content_type} = $vr->{content_type}
@@ -1958,8 +2008,7 @@ sub validate_ruleset {
 	    
 	    if ( @names_found > 1 && ! $rr->{multiple} )
 	    {
-		add_error($vr, $rr, "you may only include one of {param}",
-			  { param => \@names_found });
+		add_error($vr, $rr, 'ERR_MULT_NAMES', { param => \@names_found });
 		next RULE;
 	    }
 	    
@@ -1987,7 +2036,7 @@ sub validate_ruleset {
 	    
 	    elsif ( @raw_values > 1 && ! $rr->{multiple} )
 	    {
-		add_error($vr, $rr, "you may only specify one value for {param}: found {value}",
+		add_error($vr, $rr, 'ERR_MULT_VALUES',
 			  { param => @names_found, value => \@raw_values });
 		next RULE;
 	    }
@@ -2017,10 +2066,8 @@ sub validate_ruleset {
 	    
 	    unless ( @raw_values )
 	    {
-		add_error($vr, $rr, "you must specify a value for {param}", 
-		      { param => $rr->{param} }) if $rr->{mandatory};
-		
-		# $vr->{rs}{$ruleset_name} = 2 if $rr->{empty_ok} && ! $rr->{optional};
+		add_error($vr, $rr, 'ERR_MANDATORY', { param => $rr->{param} })
+		    if $rr->{mandatory};
 		
 		next RULE;
 	    }
@@ -2028,7 +2075,7 @@ sub validate_ruleset {
 	    # Now we process each value in turn.
 	    
 	    my @clean_values;
-	    my $error_found;
+	    my $error_flag;
 	    
 	VALUE:
 	    foreach my $raw_val ( @raw_values )
@@ -2081,17 +2128,20 @@ sub validate_ruleset {
 		    
 		    if ( $rr->{warn} )
 		    {
-			add_warning($vr, $rr, $result->{error}, { param => @names_found, value => $raw_val });
+			my $msg = $rr->{warn} ne '1' ? $rr->{warn} :
+			    $rr->{ERR_INVALID} || $rr->{errmsg} || $result->{error};
+			add_warning($vr, $rr, $msg, { param => @names_found, value => $raw_val });
 		    }
 		    
 		    # Otherwise, generate an error.
 		    
 		    else
 		    {
-			add_error($vr, $rr, $result->{error}, { param => @names_found, value => $raw_val });
+			my $msg = $rr->{ERR_INVALID} || $rr->{errmsg} || $result->{error};
+			add_error($vr, $rr, $msg, { param => @names_found, value => $raw_val });
 		    }
 		    
-		    $error_found = 1;
+		    $error_flag = 1;
 		    next VALUE;
 		}
 		
@@ -2118,43 +2168,51 @@ sub validate_ruleset {
 		push @clean_values, $value;
 	    }
 	    
-	    # If this rule allows multiple values, store a list if any
-	    # good values were found.  Otherwise, store nothing.
+	    # If clean values were found, store them.  If multiple values are
+	    # allowed, then we store them as a list.  Otherwise, there should
+	    # only be one clean value and so we just store it as a scalar.
 	    
-	    if ( $rr->{multiple} )
+	    if ( @clean_values )
 	    {
-		$vr->{clean}{$key} = \@clean_values if @clean_values;
+		if ( $rr->{multiple} )
+		{
+		    $vr->{clean}{$key} = \@clean_values;
+		}
+		
+		else
+		{
+		    $vr->{clean}{$key} = $clean_values[0];
+		}
 	    }
 	    
-	    # Otherwise, if the 'allow_multiple' setting is active, then store
-	    # a list if more than one good value was found and a single
-	    # value otherwise.
+	    # If raw values were found for this parameter, but none of them
+	    # pass the validators, then we need to indicate this condition.
 	    
-	    elsif ( $self->{SETTINGS}{allow_multiple} )
+	    else
 	    {
-		$vr->{clean}{$key} = @clean_values > 1 ? \@clean_values : $clean_values[0];
-	    }
-	    
-	    # Otherwise, if a good value was found then store it.
-	    
-	    elsif ( @clean_values )
-	    {
-		$vr->{clean}{$key} = $clean_values[0];
-	    }
-	    
-	    # If there were raw values but no clean values, and if this
-	    # rule includes a 'bad_value' key, then return its value.
-	    
-	    elsif ( @raw_values && defined $rr->{bad_value} )
-	    {
-		$vr->{clean}{$key} = $rr->{bad_value}
-		    unless defined $vr->{clean}{$key};
+		if ( defined $rr->{bad_value} && $rr->{bad_value} eq 'ERROR' )
+		{
+		    add_error($vr, $rr, 'ERR_BAD_VALUES',
+			      { param => \@names_found, value => \@raw_values });
+		    $vr->{clean}{$key} = undef;
+		    $error_flag = 1;
+		}
+		
+		elsif ( defined $rr->{bad_value} )
+		{
+		    $vr->{clean}{$key} = $rr->{multiple} ? [ $rr->{bad_value} ] : $rr->{bad_value};
+		}
+		
+		else
+		{
+		    $vr->{clean}{$key} = undef;
+		}
 	    }
 	    
 	    # Set the status of this parameter to 1 (passed) unless an error
-	    # was generated.
+	    # was generated, 0 (failed) otherwise.
 	    
-	    $vr->{ps}{$param} = $error_found ? 0 : 1;
+	    $vr->{ps}{$param} = $error_flag ? 0 : 1;
 	    
 	    # If this rule is not 'optional', then set the status of this
 	    # ruleset to 'fulfilled' (2).  That does not mean that the validation
@@ -2208,8 +2266,7 @@ sub validate_ruleset {
 	    
 	    if ( $rr->{type} eq 'together' and @present > 0 and @present < @{$rr->{param}} )
 	    {
-		add_error($vr, $rr, "you must include {param} together or not at all",
-			  { param => $rr->{param} });
+		add_error_warn($vr, $rr, 'ERR_TOGETHER', { param => $rr->{param} });
 	    }
 	    
 	    # For a 'at_most_one' rule, the count must be less than or equal
@@ -2217,8 +2274,7 @@ sub validate_ruleset {
 	    
 	    elsif ( $rr->{type} eq 'at_most_one' and @present > 1 )
 	    {
-		add_error($vr, $rr, "you may not include more than one of {param}",
-			  { param => \@present });
+		add_error_warn($vr, $rr, 'ERR_AT_MOST', { param => \@present });
 	    }
 	}
 	
@@ -2245,10 +2301,8 @@ sub validate_ruleset {
 		
 		@missing = grep { unique($_, \%found) } @{$self->{RULESETS}{$rs_name}{fulfill_order}};
 		
-		my $message = @missing == 1 ? 
-		    "you must include the parameter {param}" :
-			"you must include one of the parameters {param}";
-		add_error($vr, $rr, $message, { param => \@missing });
+		my $msg = @missing == 1 ? 'ERR_REQ_SINGLE' : 'ERR_REQ_MULT';
+		add_error_warn($vr, $rr, $msg, { param => \@missing });
 	    }
 	}
 	
@@ -2273,10 +2327,8 @@ sub validate_ruleset {
 		@missing = grep { unique($_, \%found) } 
 		    map { @{$self->{RULESETS}{$_}{fulfill_order}} } @not_fulfilled;
 		
-		my $message = @missing == 1 ? 
-		    "you must include the parameter {param}" :
-			"you must include one of the parameters {param}";
-		add_error($vr, $rr, $message, { param => \@missing });
+		my $msg = @missing == 1 ? 'ERR_REQ_SINGLE' : 'ERR_REQ_MULT';
+		add_error_warn($vr, $rr, $msg, { param => \@missing });
 	    }
 	    
 	    # For an 'allow_one' or 'require_one' rule, generate an error if
@@ -2285,15 +2337,19 @@ sub validate_ruleset {
 	    elsif ( @fulfilled > 1 and ($rr->{constraint} eq 'allow_one' or 
 					$rr->{constraint} eq 'require_one') )
 	    {
-		my (@params) = map { $self->{RULESETS}{$_}{fulfill_order} } @fulfilled;
+		my @params;
 		my ($label) = "A";
 		
-		my (@groups) = map { '(' . $label++ . ') ' . name_list(@$_) } @params;
-		my $message = @groups == 2 ?
-		    "you may not include parameters from both of these groups: {param}" :
-			"you may not include parameters from more than one of these groups: {param}";
+		foreach my $rs ( @fulfilled )
+		{
+		    push @params, "($label)"; $label++;
+		    push @params, @{$self->{RULESETS}{$rs}{fulfill_order}}
+			if ref $self->{RULESETS}{$rs}{fulfill_order} eq 'ARRAY';
+		}
 		
-		add_error($vr, $rr, $message, { param => join('; ', @groups) });
+		my $message = 'ERR_REQ_ONE';
+		
+		add_error_warn($vr, $rr, 'ERR_REQ_ONE', { param => \@params });
 	    }
 	}
 	
@@ -2319,9 +2375,7 @@ sub validate_ruleset {
 		$vr->{content_type} = 'unknown';
 		$vr->{ps}{$param} = 1;
 		$rr->{key} ||= '_content_type';
-		add_error($vr, $rr,
-			  "your request must specify a media type from the following list: {param}",
-			  { param => $rr->{type_map} });
+		add_error_warn($vr, $rr, 'ERR_MEDIA_TYPE', { param => $param, value => $rr->{type_map} });
 	    }
 	}
     }
@@ -2341,25 +2395,25 @@ sub unique {
 }
 
 
-# Add an error message to the current validation.  The @subst arguments if
-# given specify placeholder substitutions.
+# Add an error message to the current validation.
 
 sub add_error {
 
     my ($vr, $rr, $msg, $subst) = @_;
     
-    # If the rule has a 'warn' directive, add a warning instead.
+    # If no message was given, use a default one.  It's not a very good
+    # message, but what can we do?
     
-    goto &add_warning if $rr->{warn};
+    $msg ||= 'ERR_DEFAULT';
     
-    # If the rule has an 'errmsg' directive, use that message instead of the
-    # one given.
+    # If the given message starts with 'ERR_', assume it is an error code.  If
+    # the code is present as an attribute of the rule record, use the
+    # corresponding value as the message.  Otherwise, use the global value.
     
-    $msg = $rr->{errmsg} if $rr->{errmsg};
-    
-    # If no message was given, use a default one.
-    
-    $msg ||= "parameter value error";
+    if ( $msg =~ qr{^ERR_} )
+    {
+	$msg = $rr->{$msg} || $vr->{settings}{$msg} || $ERROR_MSG{$msg} || $ERROR_MSG{ERR_DEFAULT};
+    }
     
     # Next, figure out the error key.  If the rule has a 'key' directive, use
     # that.  Otherwise determine it according to the rule type, ruleset name,
@@ -2379,21 +2433,25 @@ sub add_error {
 }
 
 
-# Add a warning message to the current validation.  The @subst arguments if
-# given specify placeholder substitutions.
+# Add a warning message to the current validation.  The $subst hash if
+# given specifies placeholder substitutions.
 
 sub add_warning {
 
     my ($vr, $rr, $msg, $subst) = @_;
     
-    $msg ||= "parameter value warning";
+    # If no message was given, use a default one.  It's not a very good
+    # message, but what can we do?
     
-    # If the rule has an 'warn' directive, use that message instead of the one
-    # given.
+    $msg ||= 'ERR_DEFAULT';
     
-    if ( $rr->{warn} )
+    # If the given message starts with 'ERR_', assume it is an error code.  If
+    # the code is present as an attribute of the rule record, use the
+    # corresponding value as the message.  Otherwise, use the global value.
+    
+    if ( $msg =~ qr{^ERR_} )
     {
-	$msg = $rr->{warn} ne '1' ? $rr->{warn} : $msg;
+	$msg = $rr->{$msg} || $vr->{settings}{$msg} || $ERROR_MSG{$msg} || $ERROR_MSG{ERR_DEFAULT};
     }
     
     # Next, figure out the warning key.  If the rule has a 'key' directive, use
@@ -2413,6 +2471,29 @@ sub add_warning {
 }
 
 
+# Add an error or warning message to the current validation.  If the rule has
+# a 'warn' attribute, add a warning.  Otherwise, add an error.  If the rule
+# has an 'errmsg' attribute, use its value instead of the error message given.
+
+sub add_error_warn {
+    
+    my ($vr, $rr, $msg, $subst) = @_;
+    
+    $msg = $rr->{errmsg} if $rr->{errmsg};
+    
+    if ( $rr->{warn} )
+    {
+	$msg = $rr->{warn} if $rr->{warn} ne '1';
+	return add_warning($vr, $rr, $msg, $subst);
+    }
+    
+    else
+    {
+	return add_error($vr, $rr, $msg, $subst);
+    }
+}
+
+
 # Substitute placeholders in an error or warning message.
 
 sub subst_error {
@@ -2605,7 +2686,7 @@ sub keys {
     
     # All of the keys of the 'values' subhash represent clean parameter values.
     
-    return keys %{$self->{values}};
+    return keys %{$self->{clean}};
 }
 
 
@@ -2622,7 +2703,7 @@ sub values {
     
     # Return the clean value hash.
     
-    return $self->{values};
+    return $self->{clean};
 }
 
 =head3 value
@@ -2636,7 +2717,22 @@ sub value {
 
     my ($self, $param) = @_;
     
-    return $self->{values}{$param};
+    return $self->{clean}{$param};
+}
+
+
+=head3 specified
+
+Returns true if the specified parameter was specified in the request with at least
+one value, whether or not that value was valid.  Returns false otherwise.
+
+=cut
+
+sub specified {
+    
+    my ($self, $param) = @_;
+    
+    return exists $self->{clean}{$param};
 }
 
 
